/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 * 
 * The Original Code is Accessibar
 *
 * The Initial Developer of the Original Code is
 * Edan Kemelman <edankeme@yahoo.com>
 * Portions created by the Initial Developer are Copyright (C) 2004 -2007
 * the Initial Developer. All Rights Reserved.
 * 
 * Contributor(s):
 *    Edan Kemelman <edankeme@yahoo.com>
 *
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU Public License (the "GPL"), in which case the
 * provisions of the GPL are applicable instead of those above.
 * If you wish to allow use of your version of this file only
 * under the terms of the GPL and not to allow others to use your
 * version of this file under the MPL, indicate your decision by
 * deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL.  If you do not delete
 * the provisions above, a recipient may use your version of this
 * file under either the MPL or the GPL.
 *
 * ***** END LICENSE BLOCK ***** */

// accessibarOverlay.js contains the functions carrying out the toolbar's main tasks
// which are focused on the "visual" aspects of the toolbar. It contains the
// functions responsible for web page manipulation features and opens the various
// dialog windows (Options and About) whose functionality is then handled by separate 
// files specific to handling the functionality relevant to their own interface.
// All text-to-speech "auditory" related functionality (on the javascript side) is 
// handled in AccessibarTTS.js 


// The paths to the various style sheet files used to control web page display
const LINE_HEIGHT_NORM_CSS="chrome://accessibar/content/stylesheets/line_height_norm.css";
const LINE_HEIGHT_1_5_CSS ="chrome://accessibar/content/stylesheets/line_height_1_5.css";
const LINE_HEIGHT_2_CSS   ="chrome://accessibar/content/stylesheets/line_height_2.css";
const LINE_HEIGHT_2_5_CSS ="chrome://accessibar/content/stylesheets/line_height_2_5.css";
const LINE_HEIGHT_3_CSS   ="chrome://accessibar/content/stylesheets/line_height_3.css";
const HIDE_ALL_IMAGES_CSS ="chrome://accessibar/content/stylesheets/hide_all_images.css";

// Text filter code
const TEXT_FILTER = 1;
// Hexadecimal RGB length
const HEX_RGB_LEN = 6;

const XSI_MAX_ZOOM = 16;     // 1600% MAX for full page zoom
const XSI_MIN_ZOOM = 0.25;   // 25% MIN for full page zoom

// xsiActiveStyleIds will store id attribute strings of all styles
// to be actively applied to web document / and newly loaded documents.
var xsiActiveStyleIds;

var xsiHandledTabs = 0;


window.addEventListener("load", xsiInit, false);


function xsiAddActiveStyle(id) {        
    // Supported from FF 1.5+ 
    // Add if the style id does not already exist in the ActiveStyleIds array
    if (xsiActiveStyleIds.indexOf(id) == -1)
        xsiActiveStyleIds.push(id);
}

function xsiAddActiveStylesFromPrefs() {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
    
    if (prefs.getBoolPref("accessibar.focus.enhanceindicator")) {
           xsiAddActiveStyle("accessibar-pref-focus");
    }
    if (prefs.getBoolPref("accessibar.links.dash_underline_visited")) {
           xsiAddActiveStyle("accessibar-pref-vlnk-dashed");
    }
    // TODO: same for more pref styles that are to be applied ...
}

/**
 * This function adds a style sheet link from the given location path with the given id to
 * the supplied Array of documents. If the boolean isNew flag is true the style sheet id
 * will also be added to the list of styles currently applied.
 * @param location The path where the stylesheet to apply can be found.
 * @param id The id attribute string to set
 * @param docsArray An array of documents to apply the stylesheet to. This is needed so
 *        web pages with multiple frames can be handled properly, and can be obtained by 
 *        making a call to <code>xsiGetAllDocs()</code>.
 * @param isNew A boolean flag to indicate whether this is a new style to apply
 *        or this style has already been applied and only needs to be reapplied 
 *        to a new docsArray.
 */
function xsiAddCSSLink(location, id, docsArray, isNew)
{
    var headElem = null;
    var linkElem = null;

    // This is a new style to apply so add it to the Array of actively applied styles.
    if ( isNew ) { 
         xsiAddActiveStyle(id);
    }
    
    // Loop through all the documents to apply the style to
    for (var i = 0; i < docsArray.length; ++i) {
        headElem = docsArray[i].getElementsByTagName("head");
        
        // Create a link element that will be used to link to the style sheet.
        linkElem = docsArray[i].createElement("link");
        // These are the required attributes to link to a style sheet
        linkElem.setAttribute("href", location);
        linkElem.setAttribute("id", id);
        linkElem.setAttribute("rel", "stylesheet");
        linkElem.setAttribute("type", "text/css");
       
        // If a head element exists
        if (headElem.length > 0) {
             headElem[0].appendChild(linkElem);
        }
        else { 
             docsArray[i].documentElement.childNodes[0].appendChild(linkElem);
        }
    }
}

/**
 * Adds a CSS "style" node to the given documents.
 * Usefull for styles that are generated (such as the 70 supported color changes
 * for both bg color and fg color where it doesn't make sense to create 140 files with
 * universal css selectors "*").
 * @param styleText A string containing the CSS directives to add as "style" node child
 * @param id The id attribute string for the "style" node
 * @docsArray Array of documents to add the "style" node to.
 * @isNew Is this a new style to apply to the <code>xsiActiveStyleIds</code> array or are
 *        we reapplying a style already listed in the Array to newly loaded documents.
 */
function xsiAddCSSNode(styleText, id, docsArray, isNew)
{
    var headElem = null;
    var styleElem = null;
    var textNode = null; // Text node created to be a child of style element

    // A new style to apply to Array of actively applied styles.
    if ( isNew ) {
         xsiAddActiveStyle(id);
    }

    // Loop through all the documents to apply the style to
    for (var i = 0; i < docsArray.length; ++i) {
        headElem = docsArray[i].getElementsByTagName("head");

        styleElem = docsArray[i].createElement("style");
        styleElem.setAttribute("id", id);
        styleElem.setAttribute("type", "text/css");

        textNode = docsArray[i].createTextNode( styleText );
        
        styleElem.appendChild(textNode);

        // If a head element exists
        if (headElem.length > 0) {
            headElem[0].appendChild(styleElem);
        }
        else {
            docsArray[i].documentElement.childNodes[0].appendChild(styleElem);
        }
    }
}

/**
 * xsiApplyPrefStyles is used to apply ~preference~ styles to ALL opened TABS
 * These preference styles are: accessibar-pref-focus and accessibar-pref-vlnk-dashed
 * (Load handler will call on current page alone)
 */
function xsiApplyPrefStyles( forAllTabs ) {
    var numTabs = getBrowser().browsers.length;
    var browser;

    if (forAllTabs) {
        for (var i = 0; i < numTabs; ++i) {
            browser = getBrowser().getBrowserAtIndex(i);        // get an opened tab
            if (xsiActiveStyleIds.indexOf("accessibar-pref-focus") != -1) 
                xsiSetFocusOutline(browser.contentWindow);  // will handle all frames within the tab
            if (xsiActiveStyleIds.indexOf("accessibar-pref-vlnk-dashed") != -1) 
                xsiSetDashedVisitedLink(browser.contentWindow);  // will handle all frames within the tab
            // TODO: same for other Preference Styles...
        }
    }
    // Only apply to currently selected tab
    else {
        if (xsiActiveStyleIds.indexOf("accessibar-pref-focus") != -1) 
            xsiSetFocusOutline(window.content);  // will handle all frames within the tab
        if (xsiActiveStyleIds.indexOf("accessibar-pref-vlnk-dashed") != -1) 
            xsiSetDashedVisitedLink(window.content);  // will handle all frames within the tab
    }
}

/**
 * Applies the currently active styles. This will be called when loading a
 * web page in order to persist the styles that are registered in the 
 * <code>xsiActiveStyleIds</code> array to the newly loaded page. 
 * @param contentWindow - window.content unless were getting the contentWindow of a specific opened tab 
 *                        (not necessarily the current one) via getBrowser().contentWindow
 */
function xsiApplyStyles( contentWindow )
{
    // Documents list will be needed by xsiAddCSSLink()
    var docsArray = xsiGetAllDocs(contentWindow, new Array());
    
    for(var i = 0; i < xsiActiveStyleIds.length; ++i) {
        var aStyleId =  xsiActiveStyleIds[i];
        switch (aStyleId) {
            case "accessibar-hide-images":
            xsiAddCSSLink(HIDE_ALL_IMAGES_CSS, aStyleId, docsArray, false);
            break;
            case "accessibar-line-height-norm":
            xsiAddCSSLink(LINE_HEIGHT_NORM_CSS, aStyleId, docsArray, false);
            break;
            case "accessibar-line-height-1":
            xsiAddCSSLink(LINE_HEIGHT_1_CSS, aStyleId, docsArray, false);
            break;
            case "accessibar-line-height-1-5":
            xsiAddCSSLink(LINE_HEIGHT_1_5_CSS, aStyleId, docsArray, false);
            break;
            case "accessibar-line-height-2":
            xsiAddCSSLink(LINE_HEIGHT_2_CSS, aStyleId, docsArray, false);
            break;
            case "accessibar-line-height-2-5":
            xsiAddCSSLink(LINE_HEIGHT_2_5_CSS, aStyleId, docsArray, false);
            break;
            case "accessibar-line-height-3":
            xsiAddCSSLink(LINE_HEIGHT_3_CSS, aStyleId, docsArray, false);
            break;
        }
        if (aStyleId.indexOf("accessibar-bg-") != -1) {
             xsiReapplyColor(aStyleId, docsArray);
        }
        if (aStyleId.indexOf("accessibar-fg-") != -1) {
             xsiReapplyColor(aStyleId, docsArray);
        }
        if (aStyleId.indexOf("accessibar-lnk-") != -1) {
             xsiReapplyColor(aStyleId, docsArray);
        }
        if (aStyleId.indexOf("accessibar-vlnk-") != -1) {
             xsiReapplyColor(aStyleId, docsArray);
        }
    }
}

/**
 * This function will create a filtered document tree that can later be quickly 
 * traversed. It currently only caters to one filterType which is a TEXT_FILTER
 * but was envisioned to be expanded to handling additional filters should their
 * need arise. 
 * @param filterType The type of filter to apply when creating the tree walker
 *        at this point only TEXT_FILTER is supported, but others will be added if
 *        it should prove necessary. 
 * @param beginNode The root node of the filtered tree walker to be created.
 * @param doc The document to work with.
 * @return Returns the filtered tree walker
 */
function xsiCreateFilteredTreeWalker( filterType, beginNode, doc )
{	
    // Create a TreeWalker
    var filteredTree;
    switch(filterType) {
        case TEXT_FILTER:
        	filteredTree = doc.createTreeWalker(beginNode,
                                                    NodeFilter.SHOW_TEXT,
                                                    xsiFilterTextNodes,
                                                    false);
        break;
    }
    return filteredTree;
}

// Display the toolbar according to preferences. [Icons to hide / Display with Icons+Text or Icons]
function xsiDisplayToolbarByPrefs() 
{
    var toolbar = document.getElementById("accessibar");

    if(!toolbar || !window.toolbar.visible)
        return;
    
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
    var toolbarPref = prefs.getCharPref("accessibar.display.fullIcons");  // full [icons and text] or just icons

    toolbar.setAttribute("mode", toolbarPref);

    // Hide buttons that were selected to be hidden from the interface.
    xsiHideButtonsByPrefs();
}

/**
 * Opens the About dialog.
 */
function xsiDoAbout()
{
    window.openDialog("chrome://accessibar/content/aboutDialog.xul",
    				  "aboutDialog", "centerscreen,chrome,modal");
}

/**
 * Opens the Options dialog and once the dialog updates AccessibarTTS 
 * reader object to apply changes if it is currently active. 
 */
function xsiDoOptions()
{
    window.openDialog("chrome://accessibar/content/options.xul",
    				  "accessibar-opt-dialog", "centerscreen,chrome,toolbar,modal,resizable");
    
//	THIS CAUSES INSTABILLITY!@#!@# CHECK WHY AND THEN RESTORE WHEN RESOLVED!
//    // reader object is alive so apply changes now.
//    if ( AccessibarTTS.thisObject ) {
//    	AccessibarTTS.thisObject.setVoiceFromOptions();
//    }
}

/**
 * Defines a NodeFilter function to accept only needed text nodes which can also be used
 * as a NodeFilter function by the createTreeWaler method of the DOM API. 
 * (Skips null nodes, whitespace nodes and script nodes)
 * @param node The node to test
 * @return Returns <code>NodeFilter.FILTER_ACCEPT</code> if it is an acceptable text node
 *         and otherwise returns <code>NodeFilter.FILTER_SKIP</code>
 */
function xsiFilterTextNodes(node) 
{
	// Skip null nodes, whitespace nodes and text nodes which happen to be scripts.
    if (node.nodeValue == null ||
    	!(/[^\t\n\r\s]/.test(node.nodeValue)) || 
   		node.parentNode.nodeName == "SCRIPT") {
    	return NodeFilter.FILTER_SKIP;
    }
    return NodeFilter.FILTER_ACCEPT;
}

function xsiHideButtonsByPrefs() {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
        
    document.getElementById("textColor").collapsed = !prefs.getBoolPref("accessibar.display.txtcolor");
    document.getElementById("linkColor").collapsed = !prefs.getBoolPref("accessibar.display.lnkcolor");
    document.getElementById("BGColor").collapsed = !prefs.getBoolPref("accessibar.display.bgcolor");
    document.getElementById("xsi-fontselect").collapsed = !prefs.getBoolPref("accessibar.display.fontselect");
    document.getElementById("IncreaseFontSizeButton").collapsed = !prefs.getBoolPref("accessibar.display.txtinc");
    document.getElementById("DecreaseFontSizeButton").collapsed = !prefs.getBoolPref("accessibar.display.txtdec");
    document.getElementById("IncLineHeight").collapsed = !prefs.getBoolPref("accessibar.display.lninc");
    document.getElementById("DecLineHeight").collapsed = !prefs.getBoolPref("accessibar.display.lndec");
    document.getElementById("xsi-graphics").collapsed = !prefs.getBoolPref("accessibar.display.showhide");
    document.getElementById("xsi-read-aloud").collapsed = !prefs.getBoolPref("accessibar.display.readaloud");
    
    /** Regardless of preference - 
     ZoomIn/Out will not be displayed for unsupported versions of FF (under 3.0) */
    var ffVer = prefs.getCharPref("extensions.lastAppVersion");
    // Get the main version number as an integer
    var verInt = parseInt(ffVer.charAt(0));
    if (verInt < 3) {
        document.getElementById("ZoomInButton").collapsed = true;
        document.getElementById("ZoomOutButton").collapsed = true;
    }
    // Versions 3 and up (supporting full page zoom)
    else {
        document.getElementById("ZoomInButton").collapsed = !prefs.getBoolPref("accessibar.display.zoomin");
        document.getElementById("ZoomOutButton").collapsed = !prefs.getBoolPref("accessibar.display.zoomout");
    }
}

/** 
 * <code>xsiStartupTBarOptions()</code> is called on the initial load after being 
 * registered as a load listener by the xsiInit function. This is to ensure that it
 * is carried out ONLY AFTER the toolbar is fully loaded. Once it's been called
 * the load event listener can effectively be removed. This function's main 
 * purpose is to set the toolbar's behavior according to the various preferences.
 */
function xsiStartupTBarOptions()
{
    var prefService = Components.classes["@mozilla.org/preferences-service;1"]
                      .getService(Components.interfaces.nsIPrefService).getBranch("");

    xsiSetKbdShortcuts();
    xsiDisplayToolbarByPrefs();
    
    // If we've been called the "load" listener is no longer required
    window.removeEventListener("load", xsiStartupTBarOptions, true);

    // NEXT Update other toolbar related OPTIONs HERE..
    // Check if Reader is to be loaded
    if (prefService.prefHasUserValue("accessibar.TextToSpeech.startReaderOnLoad") &&
        prefService.getBoolPref( "accessibar.TextToSpeech.startReaderOnLoad" ) ) {
        ttsLaunchManager( true );
        document.getElementById('readerOn').setAttribute('checked', true);
        // This should be automatic considering its a radio element (but it isn't)
        document.getElementById('readerOff').setAttribute('checked', false);
    }
}

/**
* This function is responsible for setting the Keyboard Shortcuts so that they 
* reflect the values that have been saved in the user's preferences. If no such 
* value exists the user preference is set to the default key attribute obtained from
* each key element in the accessibarOverlay XUL file. 
* It calls <code>xsiSetKeyPref()</code> to carry this out this task for each one 
* of the available keyboard shortcuts keys.
*/
function xsiSetKbdShortcuts()
{
    var prefService = Components.classes["@mozilla.org/preferences-service;1"]
    .getService(Components.interfaces.nsIPrefService).getBranch("");
    
    xsiSetKeyPref("xsi-users-guide-key", "accessibar.users_guide_key", prefService);
    xsiSetKeyPref("xsi-developers-guide-key","accessibar.developers_guide_key",prefService);
    xsiSetKeyPref("xsi-options-key", "accessibar.options_key", prefService);
    xsiSetKeyPref("xsi-font-select-key", "accessibar.font_select_key", prefService);
    xsiSetKeyPref("xsi-font-zoom-in-key", "accessibar.font_zoom_in_key", prefService);
    xsiSetKeyPref("xsi-font-zoom-out-key", "accessibar.font_zoom_out_key", prefService);
    xsiSetKeyPref("xsi-space-inc-key", "accessibar.space_inc_key", prefService);
    xsiSetKeyPref("xsi-space-dec-key", "accessibar.space_dec_key", prefService);
    xsiSetKeyPref("xsi-show-graphics-key", "accessibar.show_graphics_key", prefService);
    xsiSetKeyPref("xsi-hide-graphics-key", "accessibar.hide_graphics_key", prefService);
    xsiSetKeyPref("xsi-reader-on-key", "accessibar.reader_on_key", prefService);
    xsiSetKeyPref("xsi-reader-off-key", "accessibar.reader_off_key", prefService);
    xsiSetKeyPref("xsi-undo-key", "accessibar.undo_key", prefService);
}

/**
* <code>xsiSetKeyPref</code> sets the shortcut key for a specified key element. If a 
* user preference has been set for the given key, this value is set as the key attribute
* for that element. If no preference is found in the user's preferences the default value
* is set as a char pref in the preferences service.
* @param keyToSetId The id attribute string of the key element to set.
* @param prefName The name of the preference in the user's preferences.
* @param prefService A reference to the root branch of the XPCOM nsIPrefService
*/
function xsiSetKeyPref( keyToSetId, prefName, prefService  )
{
    var key = document.getElementById( keyToSetId );
    // If the key exists
    if( key ) {
        // If the key preference is set
        if(prefService.prefHasUserValue( prefName )) {
       	     key.setAttribute("key", prefService.getCharPref( prefName ));
        }
        else {
       	     prefService.setCharPref( prefName , key.getAttribute("key"));
        }
    }
}

/**
 * <code>xsiGetAllDocs</code> takes a parent frame and stores references to all the
 * documents that this parent frame stores in its array of frames. 
 * @param frame The parent frame whose encompassing documents are needed 
 *		  (Providing <code>window.content</code> for this parameter will result
 * 		  in an Array holding all the documents of the loaded web page.
 * @param docsArray An array to store the result in.
 * @return Returns an array holding all the documents of <code>frame</code>      
 */
function xsiGetAllDocs(frame, docsArray) 
{
    // Store the current document to the given Array
    docsArray.push(frame.document);
    
    // Loop through the frames
    for(var i = 0; i < frame.frames.length; ++i) {
    	xsiGetAllDocs(frame.frames[i], docsArray);
    }
    return docsArray;
}

/**
 * Event handler handling "load" events on the currently selected page.
 * ensures styles currently stored in the xsiActiveStyleIds array are 
 * reapplied to the newly loaded page, effectively persisting the style.
 * @param event The load event to handle
 */
function xsiHandleLoad(event)
{   
    //if (!event.target || event.target != getBrowser() || !event.target.contentDocument) {
    if (!(event.originalTarget instanceof HTMLDocument) || 
        event.originalTarget.defaultView.frameElement   || 
        gBrowser.currentURI.spec.indexOf("about:blank") != -1) { 
        return;
    }  
   
    // Reapply active styles when loading a page
    if ( xsiActiveStyleIds.length > 0 ) {
    	xsiApplyStyles(window.content);   // done on the currently selected tab
        xsiApplyPrefStyles( false );      // - " -  
    }
}

/**
 * This is somewhat of a XUL UI hack to get a colorpicker to be properly accessible with the
 * keyboard when placed in a submenu. The default behavior of the left arrow keyboard event
 * is to close the menupopup upon which the colorpicker is on. What we are doing here is to prevent
 * that behavior so that selection of colors using the left arrow will be possible.
 */
function xsiHandleKeydownOnLinkCP(event){
    if (event.keyCode != 37 || event.originalTarget.id.indexOf("xsi-cp") == -1)
        return;
    
    var colorpicker = event.originalTarget;
    // we are only handling a left arrow keydown event on the Link color colorpicker
    colorpicker.parentNode.setAttribute('onpopuphiding', 'event.preventDefault(); this.setAttribute("onpopuphiding", "");');
}

/**
 * Hides all the images in all the documents in <code>window.content</code>
 */
function xsiHideImages() 
{
    var docsArray;   // Will hold an Array of the documents
    var stringsBundle = document.getElementById("accessibar-string-bundle");
    var tmpStr = null;
    var elem = null;
    
    docsArray = xsiGetAllDocs(window.content, new Array());
    
    // ** HIDE_ALL_IMAGES_CSS Style Sheet Hides FLASH Animations as well**
    xsiAddCSSLink(HIDE_ALL_IMAGES_CSS, "accessibar-hide-images", docsArray, true);
    
    // Toggle button attributes to function as restore images button
    tmpStr = stringsBundle.getString('button_graphicsPlus_accesskey');
    document.getElementById("xsi-graphics").setAttribute("accesskey", tmpStr);
    document.getElementById("xsi-graphics").setAttribute("class", "xsi-imgs-hiden");
    tmpStr = stringsBundle.getString('button_graphicsPlus_label');
    document.getElementById("xsi-graphics").setAttribute("label", tmpStr);
    elem = document.getElementById("xsi-graphics");
    elem.oncommand = xsiRestoreImages;   
    tmpStr = stringsBundle.getString('tip_graphicsPlus');
    document.getElementById("xsi-graphics").setAttribute("tooltiptext", tmpStr);    
}

/**
 * Initializes the toolbar by setting preferences and adding necessary event listeners.
 */
function xsiInit()
{
    xsiActiveStyleIds = new Array();
    
    // Called after window is loaded to set toolbar UI options and proptly removed as an event listener thereafter.
    window.addEventListener("load", xsiStartupTBarOptions, true);
    
    // colorpicker submenu left arrow fix..
    window.addEventListener("keydown", xsiHandleKeydownOnLinkCP, true);
    
    // Check prefs and add enhance-focus (and other future browse prefs) to applied pref styles when selected
    xsiAddActiveStylesFromPrefs();
    
    // A startup init to apply *style preferences* to *all* opened tabs (xsiHandleLoad will only handle currently selected tab)
    gBrowser.addEventListener("load", xsiInitPrefStylesInTabs, true);
    
    // Add event handler handling page loads
    gBrowser.addEventListener("load", xsiHandleLoad, true);
    
    gBrowser.addEventListener("load", xsiVersionCheck, true);
    
    
    try {
        window.removeEventListener("load", xsiInit, false);
    }
    catch(e) {
        Components.utils.reportError(e);
    }
}

// Used during browser initialization to apply Accessibar preference styles to all opened tabs
// ref: http://developer.mozilla.org/en/docs/Code_snippets:Tabbed_browser#Enumerating_tabs [&& #Detecting_page_load]
// - will remain a load handler until tabs are loaded - once it has handled all opened tabs, will remove itself as an event handler.
function xsiInitPrefStylesInTabs(event) {
    
    // Treat the root HTMLDocument and not internal frames if they exist (which will also fire load events) 
    if (event.originalTarget instanceof HTMLDocument && !event.originalTarget.defaultView.frameElement && gBrowser.currentURI.spec.indexOf("about:blank") == -1) {
        var numTabs = gBrowser.browsers.length;
        var currDocIndx = gBrowser.getBrowserIndexForDocument( event.originalTarget );
        var browser = gBrowser.getBrowserAtIndex( currDocIndx );    // get current tab loading
        
        // apply accessibar-pref-focus preference
        if (xsiActiveStyleIds.indexOf("accessibar-pref-focus") != -1)
            xsiSetFocusOutline(browser.contentWindow);  // will handle all frames within the tab
        if (xsiActiveStyleIds.indexOf("accessibar-pref-vlnk-dashed") != -1) 
            xsiSetDashedVisitedLink(browser.contentWindow);  // will handle all frames within the tab       
        // apply others...           
        
        xsiHandledTabs++;

        // done loading all tabs (...except about:blank,config etc.) - can now remove listener
        if (xsiHandledTabs == numTabs) {
            gBrowser.removeEventListener("load", xsiInitPrefStylesInTabs, true); 
        }
    }
}

/**
* Tests whether the given node is a valid text node.
* @param node The node to test
* @return Returns true if tested node is valid and false otherwise.
*/
function xsiIsValidTextNode(node) 
{
    return (node.nodeType == Node.TEXT_NODE && 
            xsiFilterTextNodes(node) == NodeFilter.FILTER_ACCEPT);
}

/**
 * This function decreases the line height by applying the appropriate style sheet
 * to all the loaded web page documents.
 */
function xsiLineHeightDec()
{		
    var docsArray;			  // Will hold an Array of the documents
    var isApplied = false;    // A line-height style is currently applied
    
    docsArray = xsiGetAllDocs(window.content, new Array());
    
    // Check to see if one of the style sheets is applied. If so, remove the applied sheet
    // and apply the sheet with the next smaller value of the previously applied one.
    for (var i=0; i < xsiActiveStyleIds.length; ++i) {
        switch ( xsiActiveStyleIds[i] ) {
            // If the minimum sized line-height is applied we just flag it as applied
            case "accessibar-line-height-norm": isApplied = true;    break;
            case "accessibar-line-height-1-5": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_NORM_CSS,"accessibar-line-height-norm",docsArray, true);
            xsiRemoveCSS("accessibar-line-height-1-5", docsArray);
            break;
            case "accessibar-line-height-2": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_1_5_CSS, "accessibar-line-height-1-5", docsArray, true);
            xsiRemoveCSS("accessibar-line-height-2", docsArray);
            break;
            case "accessibar-line-height-2-5": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_2_CSS, "accessibar-line-height-2", docsArray, true);
            xsiRemoveCSS("accessibar-line-height-2-5", docsArray);
            break;
            case "accessibar-line-height-3": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_2_5_CSS, "accessibar-line-height-2-5", docsArray, true);
            xsiRemoveCSS("accessibar-line-height-3", docsArray);
            break;
        }
        if (isApplied) {
        	break;
        }
    }
}

/**
* This function increases the line height on the displayed HTML documents.
* This is done by applying a relevant style sheet each time.
*/
function xsiLineHeightInc()
{
    var docsArray;
    var isApplied = false;    // A line-height style is currently applied
    
    docsArray = xsiGetAllDocs(window.content, new Array());
    
    // Check to see if one of the style sheets is applied. If so, remove the applied sheet
    // and apply the sheet with the next largest value of the previously applied one...
    // Otherwise we just add the second largest one.
    for (var i=0; i < xsiActiveStyleIds.length; ++i) {
        // New CSS is added first followed by the removing of the old one in order to
        // avoid the big changes while it is removed and until the new one is applied..
        switch ( xsiActiveStyleIds[i] ) {
            // The new style is added prior to removing the old so that there won't be
            // a jumpy intermittent state where the height returns to the original author
            // setting. The temporary conflict of having two conflicting CSS properties is
            // resolved by the order applied in which the later style prevails.
            case "accessibar-line-height-norm": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_1_5_CSS, "accessibar-line-height-1-5", docsArray, true);
            xsiRemoveCSS("accessibar-line-height-norm", docsArray);
            break;
            case "accessibar-line-height-1-5": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_2_CSS, "accessibar-line-height-2", docsArray, true);
            xsiRemoveCSS("accessibar-line-height-1-5", docsArray);
            break;
            case "accessibar-line-height-2": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_2_5_CSS, "accessibar-line-height-2-5", docsArray, true);
            xsiRemoveCSS("accessibar-line-height-2", docsArray);
            break;
            case "accessibar-line-height-2-5": isApplied = true;
            xsiAddCSSLink(LINE_HEIGHT_3_CSS, "accessibar-line-height-3", docsArray, true);
            xsiRemoveCSS("accessibar-line-height-2-5", docsArray);
            break;
            // If the maximum sized line-height is applied we just flag it as applied
            case "accessibar-line-height-3": isApplied = true;    break;
        }
        if (isApplied) {
        	break;
        }
    }
    // No previous line-height CSS has been applied so we add the second largest one.
    if (!isApplied) {
    	xsiAddCSSLink(LINE_HEIGHT_1_5_CSS, "accessibar-line-height-1-5", docsArray, true);
    }
}

/**
 * Loads the given url as the current displayed document
 * @param url The url to load
 */
function xsiLoadPage(url)
{
    window.content.document.location.href=url;
}

/**
 * Open current toolbarbutton when hovering, closing others 
 */
function xsiOpenTbBtn(currBtn)
{
    if(currBtn && !currBtn.open) {
        var tbBtns      = currBtn.parentNode.getElementsByTagName("toolbarbutton");
        var tbBtnsLen   = tbBtns.length;
        
        for(var i = 0; i < tbBtnsLen; i++) {
            tbBtn = tbBtns.item(i);
            if(tbBtns.item(i) && tbBtns.item(i) != currBtn && tbBtns.item(i).open) {
                tbBtns.item(i).open = false;
                currBtn.open = true;
                break;
            }
        }
    }
}

/**
 * This function is used to reapply a color style already listed in our array of
 * active styles. We extract the color RGB from the Id attribute string of the style.
 * @param id The id attribute of the style
 * @param docsArray Array of documents to reapply color to.
 */
function xsiReapplyColor(id, docsArray)
{
     // uses xsiAddCSSNODE
     var styleText;
     var theColor;

     // A Background style node is to be reapplied
     if (id.indexOf("accessibar-bg-") != -1) {
          // Extract the color from its id string
          theColor = "#" + id.substring(id.length-HEX_RGB_LEN, id.length);
          styleText = "* { background-color: " + theColor + " !important; }";
                     // + " background-image: none !important; }";
     }
     // A Foreground style node is to be reapplied
     else if (id.indexOf("accessibar-fg-") != -1) {
          theColor = "#" + id.substring(id.length-HEX_RGB_LEN, id.length);
          styleText = "* { color: " + theColor + " !important; }";
     }
     // A hyperlink style node is to be reapplied
     else if (id.indexOf("accessibar-lnk-") != -1) {
          theColor = "#" + id.substring(id.length-HEX_RGB_LEN, id.length);
          styleText = "a { color: " + theColor + " !important; }";
     }
     // A visited hyperlink style node is to be reapplied
     else if (id.indexOf("accessibar-vlnk-") != -1) {
          theColor = "#" + id.substring(id.length-HEX_RGB_LEN, id.length);
          styleText = "a:visited { color: " + theColor + " !important; }";
     }
     xsiAddCSSNode(styleText, id, docsArray, false);
}

/**
 * <code>xsiRemoveActiveStyle</code> Removes an actively applied style entry from the
 * <code>xsiActiveStyleIds</code> array.
 * @param id The id attribute string of the style to be removed from the array of active
 *        styles 
 */
function xsiRemoveActiveStyle(id)
{
    // Loop through the actively applied styles
    for(var i = 0; i < xsiActiveStyleIds.length; ++i) {
        // If this style sheet is in the active styles
        if(xsiActiveStyleIds[i] == id) {
       	     xsiActiveStyleIds.splice(i, 1);
        }
    }
}

/**
 * This function removes the style sheet with the given id.
 * @param id The is attribute string of the style to remove
 * @param docsArray An array of documents to remove the style from.
 */
function xsiRemoveCSS(id, docsArray)
{
    var element = null;
    
    // Loop through the documents
    for(var i = 0; i < docsArray.length; ++i) {
        // Remove all elements with that id
        while(element = docsArray[i].getElementById(id)) {
             element.parentNode.removeChild(element);
        }
    }
    xsiRemoveActiveStyle(id);
}

/**
 * Since ~pref styles~ (i.e.: accessibar-pref-XXX) are applied to ALL opened tabs, we use
 * this function to remove the applied style with the given id from all the opened tabs.
 */
function xsiRemovePrefStyle(id) {
    var numTabs = getBrowser().browsers.length;
    var browser;
    var docsArray;
    
    // Run xsiRemoveCSS on every opened tab.
    for (var i = 0; i < numTabs; ++i) {
        browser = getBrowser().getBrowserAtIndex(i);        // get an opened tab
        docsArray = xsiGetAllDocs( browser.contentWindow, new Array() );

        xsiRemoveCSS(id, docsArray);  
    }
}

/**
 * Reparents a given node with a new parent of a given type.
 * @param node The node to reparent
 * @param newParentType A string holding the name of the element type to create as a
 *        new parent for the given node.
 */
function xsiReparentNode(node, newParentType) 
{
	// Create a new parent element
    var newParent = window.content.document.createElement(newParentType); 
    // Get the parent of the node
    var oldParent = node.parentNode; 
    // Replace the node with the new parent                      
    oldParent.replaceChild(newParent, node);
    // Make the node a child of the new parent              
    newParent.appendChild(node);                           
    
    return newParent;
}

/**
 * Restores images that have been hidden using the <code>xsiHideImages()</code> function. 
 */
function xsiRestoreImages() 
{
    var docsArray = xsiGetAllDocs( window.content, new Array() );
    var stringsBundle = document.getElementById("accessibar-string-bundle");
    var tmpStr = null;
    var elem = null;
    
    xsiRemoveCSS('accessibar-hide-images', docsArray);
    
    // Toggle button attributes to function as hide images button
    tmpStr = stringsBundle.getString('button_graphicsMinus_accesskey');
    document.getElementById("xsi-graphics").setAttribute("accesskey", tmpStr);
    document.getElementById("xsi-graphics").setAttribute("class", "xsi-imgs-shown");
    tmpStr = stringsBundle.getString('button_graphicsMinus_label');
    document.getElementById("xsi-graphics").setAttribute("label", tmpStr);
    elem =  document.getElementById("xsi-graphics");
    elem.oncommand = xsiHideImages;   
    tmpStr = stringsBundle.getString('tip_graphicsMinus');
    document.getElementById("xsi-graphics").setAttribute("tooltiptext", tmpStr);    
}

/**
 * Sets a background or foreground color. Instead of using hundreds of style sheets having
 * the Universal CSS Selector "*" and the specific color, we generate a style string to
 * be added to the documents using <code>xsiAddCSSNode()</code>.
 * @param colorPicker The colorpicker XUL element used to select the color
 * @param setWhat - "bg" sets background color, "fg" sets text color, "lnk" sets hyperlink color
 */
function xsiSetColor( colorPicker, setWhat )
{
    var color = colorPicker.mSelectedCell.getAttribute('color');
    var docsArray = xsiGetAllDocs(window.content, new Array());
    var styleText;
    var styleId;       // A Generated style id attribute for "style" node
    
    if ( setWhat == "bg" ) {
        // Remove accessibar bg color style already applied to page
        for (var i=0; i < xsiActiveStyleIds.length; ++i) {
            // We've got an accessibar-bg style in our active styles array to remove
            if( xsiActiveStyleIds[i].indexOf("accessibar-bg-") != -1) {
                 xsiRemoveCSS(xsiActiveStyleIds[i], docsArray);
                 break;   // Shouldn't be more
            }
        }
        styleText = "* { background-color: " + color + " !important; }";
                  // + " background-image: none !important; }";
        // Generate an Id attribute to use for Style element, discarding color's
        // beginning # sign (to adhere to w3c standards for legal id attribute characters)
        styleId = "accessibar-bg-" + color.substring(1, color.length);
    }
    else if (setWhat == "fg") {
        // Remove accessibar fg color style already applied to page
        for (var i=0; i < xsiActiveStyleIds.length; ++i) {
            // Remove accessibar-fg style from our active styles array
            if( xsiActiveStyleIds[i].indexOf("accessibar-fg-") != -1) {
                 xsiRemoveCSS(xsiActiveStyleIds[i], docsArray);
                 break;
            }
        }
        styleText = "* { color: " + color + " !important; }";
        styleId = "accessibar-fg-" + color.substring(1, color.length);
    }
    else if (setWhat == "lnk"){
         for (var i=0; i < xsiActiveStyleIds.length; ++i) {
            if( xsiActiveStyleIds[i].indexOf("accessibar-lnk-") != -1) {
                 xsiRemoveCSS(xsiActiveStyleIds[i], docsArray);
                 break;
            }
        }
        styleText = "a { color: " + color + " !important; }";
        styleId = "accessibar-lnk-" + color.substring(1, color.length);
    }
    else if (setWhat == "vlnk"){
         for (var i=0; i < xsiActiveStyleIds.length; ++i) {
            if( xsiActiveStyleIds[i].indexOf("accessibar-vlnk-") != -1) {
                 xsiRemoveCSS(xsiActiveStyleIds[i], docsArray);
                 break;
            }
        }
        styleText = "a:visited { color: " + color + " !important; }";
        styleId = "accessibar-vlnk-" + color.substring(1, color.length);
    }
    xsiAddCSSNode(styleText, styleId, docsArray, true);
}

function xsiSetDashedVisitedLink( contentWindow ){
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
    var docsArray = xsiGetAllDocs(contentWindow, new Array());
    var isInAt = xsiActiveStyleIds.indexOf("accessibar-pref-vlnk-dashed"); // style index in active style array
   
    // We've got an existing accessibar-pref-focus style in our active pref styles array - so remove before updating
    if( isInAt != -1) 
         xsiRemoveCSS(xsiActiveStyleIds[isInAt], docsArray);
    
    var styleText = 'a:visited { text-decoration: none !important; border-bottom: dashed 0.1em !important; }';
    
    xsiAddCSSNode(styleText, "accessibar-pref-vlnk-dashed", docsArray, true);
}

/**
 * Creating a style node to enhance the way Firefox's focus indicator is presented to users. 
 * (Process somewhat similar to xsiSetColor() - may wish to generalize shared section as a separate function)
 * will be obtaining the color/width from prefs
 */
function xsiSetFocusOutline( contentWindow ) {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
    var focusColor = prefs.getCharPref("accessibar.focus.color");
    var focusWidth = prefs.getCharPref("accessibar.focus.width");
    var docsArray = xsiGetAllDocs(contentWindow, new Array());
     
    var isInAt = xsiActiveStyleIds.indexOf("accessibar-pref-focus"); // Focus style index in active style array
   
    // We've got an existing accessibar-pref-focus style in our active pref styles array - so remove before updating
    if( isInAt != -1) 
         xsiRemoveCSS(xsiActiveStyleIds[isInAt], docsArray);
           
    // may prefer to read this from a file and do text replacment..
    var styleText = '*:focus { -moz-outline: ' + focusWidth + ' solid ' + focusColor + '; -moz-outline-offset: 1px; -moz-outline-radius: 5px; }\n' +
              'button:focus::-moz-focus-inner { border-color: transparent; }\n' +
              'button::-moz-focus-inner,\n' +
              'input[type="reset"]::-moz-focus-inner,\n' +
              'input[type="button"]::-moz-focus-inner,\n' +
              'input[type="submit"]::-moz-focus-inner,\n' +
              'input[type="file"] > input[type="button"]::-moz-focus-inner {\n' +
              'border: 1px dotted transparent;\n' +
              '}\n' +
              'textarea:focus, button:focus, select:focus, input:focus { -moz-outline-offset: -1px; }\n' +
              'input[type="radio"]:focus {-moz-outline-radius: 12px; -moz-outline-offset: 0px; }\n' +
              'a:focus { -moz-outline-offset: 0px; }';

    xsiAddCSSNode(styleText, "accessibar-pref-focus", docsArray, true);
}

/**
 * Performs necessary operations needed before shutting down the browser.
 * Currently the background and foreground user preference colors are
 * restored to their original values.
 * (THIS IS COMMENTED OUT SINCE COLORS ARE NO LONGER CHANGED USING THE PREF SERVICE,
 *  BUT OTHER SHUTDOWN REQUIREMENTS MIGHT ARISE...)
 *
function xsiShutdown() 
{
    var nsIPref = Components.classes["@mozilla.org/preferences-service;1"]
                  .getService(Components.interfaces.nsIPrefService);
    var prefBranch = nsIPref.getBranch("");
    
    // restore original browser color settings
    prefBranch.setCharPref("browser.display.foreground_color", xsiOrigFG);
    prefBranch.setCharPref("browser.display.background_color", xsiOrigBG);
    prefBranch.setBoolPref('browser.display.use_document_colors', true);
}
*/

/**
 * <code>xsiTraverseTreeWalker</code> accepts a TreeWalker Object, and a function to be
 * applied on each one of the nodes of the tree.  A filtered tree walker can be created 
 * by the <code>xsiCreateFilteredTreeWalker()</code> function.
 * @param walker The tree walker
 * @param applyFunc A function to be applied to each node in the tree. 
 */
function xsiTraverseTreeWalker( walker, applyFunc ) 
{
    if (walker.firstChild()) {
        do {
            if (applyFunc != null) {
                applyFunc(walker.currentNode);
            }
            
            xsiTraverseTreeWalker( walker, applyFunc );
            
        } while(walker.nextSibling());
        // don't forget to return the treewalker to it's previous state
        // before exiting the function
        walker.parentNode();
    }
}

/**
 * This function will restore the web page to its original state.
 * All settings will be reapplied to all newly loaded pages until this function is called.
 */
function xsiUndoSettings()
{
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");

    // remove all active style sheet ids
    xsiActiveStyleIds.splice(0);

    // If user set accessibar.focus.enhanceindicator pref to true restore it as an active style
    if (prefs.getBoolPref("accessibar.focus.enhanceindicator"))
        xsiAddActiveStyle("accessibar-pref-focus");
    if (prefs.getBoolPref("accessibar.links.dash_underline_visited"))
        xsiAddActiveStyle("accessibar-pref-vlnk-dashed");


    //(artifact - no longer change colors with pref service)
    // restore original browser color settings
    //prefBranch.setCharPref("browser.display.foreground_color", "#000000");
    //prefBranch.setCharPref("browser.display.background_color", "#ffffff");
    //prefBranch.setBoolPref('browser.display.use_document_colors', true);
    
    // Update page now so changes can take effect.
    BrowserReload();
    
    // Graphics were previously hidden and are no longer, so enable 'Hide-Graphics'
    if (document.getElementById("xsi-graphics").getAttribute("class") == "xsi-imgs-hiden") {
        // will change xsi-graphics toolbarbutton to function as Hide-Graphics btn 
        xsiRestoreImages();
    }
    
      // reset the text size
    ZoomManager.prototype.getInstance().reset();
    
    // reset Full Page Zoom value
    getBrowser().mCurrentBrowser.markupDocumentViewer.fullZoom = 1;
}

function xsiVersionCheck(){
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
    var stringsBundle = document.getElementById("accessibar-string-bundle");
    var thisVer;
    
    gBrowser.removeEventListener("load", xsiVersionCheck, true);

    // This is the first time the user has installed Accessibar
    if (prefs.getCharPref("accessibar.version") == "0.0") {
        thisVer = stringsBundle.getString('accessibar.version');
        // Set the current version in prefs
        prefs.setCharPref("accessibar.version", thisVer);

        gBrowser.selectedTab = gBrowser.addTab("http://www.accessibar.com/0.7.8/releasenotes/index.html");
    }

}

/**
 *  Integrating interface for FF3's new Full Page Zoom:
 *  Adapted from Ted Mielczarek's Full Page Zoom extension ( http://developer.mozilla.org/en/docs/Full_page_zoom )
 */

 //  RESET CALL IS ==>    getBrowser().mCurrentBrowser.markupDocumentViewer.fullZoom = 1;

function xsiZoomFactorCalc() {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("");
    return prefs.getIntPref("accessibar.fullpagezoom.zoomfactor") / 100;
}

// Zooming in enforcing a reasonable upper end value
function xsiZoomIn() {
    var currZoom = getBrowser().mCurrentBrowser.markupDocumentViewer.fullZoom;

    if (currZoom >= XSI_MAX_ZOOM)
       return;

    currZoom *= xsiZoomFactorCalc();
    getBrowser().mCurrentBrowser.markupDocumentViewer.fullZoom = currZoom;
}

// Zooming out whilst enforcing a reasonable lower end value
function xsiZoomOut() {
    var currZoom = getBrowser().mCurrentBrowser.markupDocumentViewer.fullZoom;

    if (currZoom <= XSI_MIN_ZOOM)
       return;

    currZoom /= xsiZoomFactorCalc();
    getBrowser().mCurrentBrowser.markupDocumentViewer.fullZoom = currZoom;
}

